/*
 * rgb133kernel.c
 *
 * Copyright (c) 2009 Datapath Ltd. All rights reserved.
 *
 * All information contained herein is proprietary and
 * confidential to Datapath Ltd. and is licensed under
 * the terms of the Datapath Ltd. Software License.
 * Please read the LICENCE file for full license terms
 * and conditions.
 *
 * http://www.datapath.co.uk/
 * support@datapath.co.uk
 *
 */

#include <linux/kernel.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/workqueue.h>
#include <linux/firmware.h>
#include <linux/pci.h>
#include <linux/device.h>

#include <asm/uaccess.h>

#include "rgb133config.h"
#include "rgb133.h"
#include "rgb133deviceapi.h"
#include "rgb133kernel.h"
#include "rgb133debug.h"

bool KernelUseWorkLookup(void)
{
#ifndef RGB133_CONFIG_HAVE_NEW_WORK
   return FALSE;
#else
   return TRUE;
#endif
}

void* KernelVzalloc(unsigned int size)
{
   void* ptr = vmalloc(size);
   if(ptr)
      memset(ptr, 0, size);
   else
   {
      RGB133PRINT((RGB133_LINUX_DBG_ERROR, "KernelVzalloc: failed to alloc %d bytes\n",
            size));
   }
   return ptr;
}

void KernelVfree(void* ptr)
{
   vfree(ptr);
}

void* KernelKmalloc(size_t size, int flags)
{
   void* ptr = kmalloc(size, flags);
   if(ptr)
      memset(ptr, 0, size);
   else
   {
      RGB133PRINT((RGB133_LINUX_DBG_ERROR, "KernelKmalloc: failed to alloc %d bytes\n",
            size));
   }
   return ptr;
}

void KernelKfree(const void* ptr)
{
   kfree(ptr);
}

void KernelInitWork(void* arg1, void* arg2, void* arg3)
{
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,18))
   //RGB133PRINT((RGB133_LINUX_DBG_TODO, "KernelInitWork: INIT_WORK(0x%p) - data(0x%x)\n", arg1, arg3));
   INIT_WORK((struct work_struct*)arg1, arg2, arg3);
#else
   //RGB133PRINT((RGB133_LINUX_DBG_TODO, "KernelInitWork: INIT_WORK(0x%p)\n", arg1));
   INIT_WORK((struct work_struct*)arg1, arg2);
#endif /* RGB133PRE */
}

void KernelPrepareWork(void* arg1, void* arg2, void* arg3)
{
#ifdef RGB133_CONFIG_HAVE_MACRO_PREPARE_WORK
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,18))
   //RGB133PRINT((RGB133_LINUX_DBG_TODO, "KernelPrepareWork: PREPARE_WORK(0x%p) - data(0x%x)\n", arg1, arg3));
   PREPARE_WORK((struct work_struct*)arg1, arg2, arg3);
#else
   //RGB133PRINT((RGB133_LINUX_DBG_TODO, "KernelPrepareWork: PREPARE_WORK(0x%p)\n", arg1));
   PREPARE_WORK((struct work_struct*)arg1, arg2);
#endif /* RGB133PRE */
#else /* PREPARE_WORK does not exist */
   struct work_struct* ws = (struct work_struct*)arg1;
   ws->func = arg2;
#endif /* PREPARE_WORK */
}

int KernelLoadFirmware(const PFWENTRYAPI* pfw_entry, char* fw_name, PDEAPI _pDE)
{
   const struct firmware** fw_entry = (const struct firmware**)pfw_entry;
   struct rgb133_dev* rgb133 = DeviceGetRGB133(_pDE);

   if(rgb133)
      return request_firmware(fw_entry, fw_name, (struct device*)KernelGetDriverDevice(rgb133));
   else
      return 0;
}

int KernelGetFwSize(PFWENTRYAPI fw_entry)
{
   return ((struct firmware*)fw_entry)->size;
}

char* KernelGetFwData(PFWENTRYAPI fw_entry)
{
   return (char*)((struct firmware*)fw_entry)->data;
}

void KernelReleaseFw(PFWENTRYAPI fw_entry)
{
   release_firmware((struct firmware*)fw_entry);
}

PDRIVERDEVAPI KernelGetDriverDevice(struct rgb133_dev* rgb133)
{
   return (PDRIVERDEVAPI)&rgb133->core.pci->dev;
}

void KernelAtomicSet(atomic_t* pAtomic, int value)
{
   atomic_set(pAtomic, value);
}

int KernelAtomicRead(const atomic_t* pAtomic)
{
   return atomic_read(pAtomic);
}

void KernelDoGettimeofday(unsigned long* pSec, unsigned long* pUsec)
{
   struct timeval tv;
   do_gettimeofday(&tv);
   *pSec = tv.tv_sec;
   *pUsec = tv.tv_usec;
}

int KernelMemcmp(const void* s1, const void* s2, size_t n)
{
   return memcmp(s1, s2, n);
}

int KernelCopyFromUser(void* to, const __user void* from, unsigned long count)
{
   return copy_from_user(to, from, count);
}

int KernelCopyToUser(__user void* to, void* from, unsigned long count)
{
   return copy_to_user(to, from, count);
}

void KernelMutexInit(PMUTEXAPI pLock)
{
   mutex_init((struct mutex*)pLock);
}

void KernelMutexLock(PMUTEXAPI pLock)
{
   mutex_lock((struct mutex*)pLock);
}

void KernelMutexUnlock(PMUTEXAPI pLock)
{
   mutex_unlock((struct mutex*)pLock);
}

void KernelPreemptEnable( )
{
   preempt_enable();
}

void KernelPreemptDisable( )
{
   preempt_disable();
}

void KernelMemset(void* ptr, int value, unsigned int size)
{
   if(ptr)
      memset(ptr, value, size);
   else
      RGB133PRINT((RGB133_LINUX_DBG_WARNING, "KernelMemset: argument is a NULL pointer\n"));
}

void KernelAddTimer(PTIMERAPI _pTimer)
{
   struct timer_list* pTimer = (struct timer_list*)_pTimer;

   add_timer(pTimer);
}

int KernelDelTimer(PTIMERAPI _pTimer)
{
   struct timer_list* pTimer = (struct timer_list*)_pTimer;

   return del_timer(pTimer);
}

BOOL KernelInitAndAddTimer(PHANDLE pTimerAudio, PVOID start_routine, PVOID Data)
{
   // Allocate memory for the timer
   struct timer_list* pTimer = (struct timer_list*)KernelVzalloc(sizeof(struct timer_list));

   // step 1: Initialising function for the timer must be called before any other timer functions are called
   init_timer(pTimer);
   // step 2: set timer fields as needed
   pTimer->data = (unsigned long)Data;
   pTimer->function = ((void(*)(unsigned long))start_routine);
   pTimer->expires = jiffies + 1;
   *pTimerAudio = (HANDLE) pTimer;

   // step 3: register timer
   add_timer((struct timer_list*)(*pTimerAudio));

   return TRUE;
}

void KernelSetTimerExpires(PTIMERAPI _pTimer, BOOL use_jiffies, unsigned long value_jffs)
{
   struct timer_list* pTimer = (struct timer_list*)_pTimer;

   if(use_jiffies)
      pTimer->expires = jiffies + value_jffs;
   else
      pTimer->expires = value_jffs;
}

PMODULEAPI KernelGetModule( )
{
   return (PMODULEAPI)THIS_MODULE;
}

int KernelGetGFP_DMAFlag( )
{
   return GFP_DMA;
}

unsigned long KernelGetPAGE_OFFSET( )
{
   return PAGE_OFFSET;
}

unsigned int KernelGetDeviceNodeNr(void* _dev, int channel)
{
   struct rgb133_dev* dev = (struct rgb133_dev*)_dev;
   unsigned int node = -1;

   /* If inputs are exposed, we create a video_device per channel so we can dereference device info map by channel number
    * Otherwise, the device we are looking for is always at index 0
    */
   if(dev->devices <= 1)
      channel = 0;

#ifdef RGB133_CONFIG_HAVE_NEW_VDEV_NUM
   node = dev->pVfd[channel]->num;
#else
   node = dev->pVfd[channel]->minor;
#endif

   return node;
}



